//
// FM synthesizer
//

module fms(clk, rst, start,
           modFrequ, modDecay,
           carFrequ, carDecay,
           sample);
    input clk;
    input rst;
    input start;
    input [15:0] modFrequ;
    input [15:0] modDecay;
    input [15:0] carFrequ;
    input [15:0] carDecay;
    output reg [11:0] sample;

  reg [11:0] clockCount;
  wire sampleTick;
  reg [7:0] sampleCount;
  wire controlTick;

  reg [4:0] state;
  reg [4:0] nextState;
  reg [2:0] amux;
  reg [2:0] bmux;
  reg [2:0] zmux;

  reg [15:0] a;
  reg [15:0] b;
  reg [15:0] z;

  wire [15:0] addValue;
  wire [15:0] mulValue;
  wire [15:0] sineValue;
  wire [15:0] afract;
  wire [15:0] expValue;

  reg [15:0] modOscDelta;
  reg [15:0] modEnvDelta;
  reg [15:0] carOscDelta;
  reg [15:0] carEnvDelta;

  reg [15:0] modOscPhase;
  reg loadModOscPhase;
  reg [15:0] modEnvPhase;
  reg loadModEnvPhase;
  reg [15:0] carOscPhase;
  reg loadCarOscPhase;
  reg [15:0] carEnvPhase;
  reg loadCarEnvPhase;
  reg [15:0] aux1;
  reg loadAux1;
  reg [15:0] aux2;
  reg loadAux2;
  reg [3:0] mulShift;
  reg loadMulShift;
  reg [19:0] signedTotal;
  reg clearSignedTotal;
  reg addToSignedTotal;
  wire [19:0] unsignedTotal;
  reg loadSample;

  //
  // start a new note
  //

  always @(posedge clk) begin
    if (rst) begin
      modOscDelta <= 0;
      modEnvDelta <= 0;
      carOscDelta <= 0;
      carEnvDelta <= 0;
    end else begin
      if (start) begin
        modOscDelta <= modFrequ;
        modEnvDelta <= modDecay;
        carOscDelta <= carFrequ;
        carEnvDelta <= carDecay;
      end
    end
  end

  //
  // generate sample and control ticks
  //
  // sampleTick frequency: fs = 100MHz / 4096
  // sampleTick duration: ds = 1 / 100MHz
  //
  // controlTick frequency: fc = fs / 256
  // controlTick duration: dc = 1 / fs
  //

  always @(posedge clk) begin
    if (rst) begin
      clockCount <= 12'h000;
    end else begin
      clockCount <= clockCount + 1;
    end
  end

  assign sampleTick = (clockCount == 12'h000 ? 1 : 0);

  always @(posedge clk) begin
    if (rst | start) begin
      sampleCount <= 8'h00;
    end else begin
      if (sampleTick) begin
        sampleCount <= sampleCount + 1;
      end
    end
  end

  assign controlTick = (sampleCount == 8'h00 ? 1 : 0);

  //
  // the synthesizer's fsm
  //

  always @(posedge clk) begin
    if (rst | start) begin
      state <= 5'd0;
    end else begin
      state <= nextState;
    end
  end

  always @(*) begin
    case (state)
      5'd0:     begin
                  // wait for next sample tick
                  // signedTotal <- 0
                  if (sampleTick) begin
                    nextState = 5'd1;
                  end else begin
                    nextState = 5'd0;
                  end
                  amux = 3'bxxx;
                  bmux = 3'bxxx;
                  zmux = 3'bxxx;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b1;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd1:     begin
                  // aux1 <- modOscPhase
                  nextState = 5'd2;
                  amux = 3'b000;
                  bmux = 3'bxxx;
                  zmux = 3'b000;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b1;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd2:     begin
                  // modOscPhase <- modOscPhase + modOscDelta
                  nextState = 5'd3;
                  amux = 3'b000;
                  bmux = 3'b000;
                  zmux = 3'b010;
                  loadModOscPhase = 1'b1;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd3:     begin
                  // aux1 <- sine(aux1) (1)
                  nextState = 5'd4;
                  amux = 3'b100;
                  bmux = 3'bxxx;
                  zmux = 3'bxxx;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd4:     begin
                  // aux1 <- sine(aux1) (2)
                  nextState = 5'd5;
                  amux = 3'bxxx;
                  bmux = 3'bxxx;
                  zmux = 3'b100;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b1;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd5:     begin
                  // aux2 <- modEnvPhase
                  nextState = 5'd6;
                  amux = 3'b001;
                  bmux = 3'bxxx;
                  zmux = 3'b000;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b1;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd6:     begin
                  // modEnvPhase <- modEnvPhase + modEnvDelta
                  // (but only if controlTick is active AND
                  // integer part of envelope phase < 15)
                  nextState = 5'd7;
                  amux = 3'b001;
                  bmux = 3'b001;
                  zmux = 3'b010;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = controlTick &
                                    (a[15:12] != 4'd15);
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd7:     begin
                  // aux2 <- exp(fract(aux2)) (1)
                  // mulShift <- integ(aux2)
                  nextState = 5'd8;
                  amux = 3'b101;
                  bmux = 3'bxxx;
                  zmux = 3'b000;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b1;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd8:     begin
                  // aux2 <- exp(fract(aux2)) (2)
                  nextState = 5'd9;
                  amux = 3'bxxx;
                  bmux = 3'bxxx;
                  zmux = 3'b101;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b1;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd9:     begin
                  // aux1 <- (aux1 * aux2) >> (mulShift + 15) (1)
                  nextState = 5'd10;
                  amux = 3'b100;
                  bmux = 3'b101;
                  zmux = 3'bxxx;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd10:    begin
                  // aux1 <- (aux1 * aux2) >> (mulShift + 15) (2)
                  nextState = 5'd11;
                  amux = 3'bxxx;
                  bmux = 3'bxxx;
                  zmux = 3'b011;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b1;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd11:    begin
                  // aux2 <- aux1 + aux1
                  nextState = 5'd12;
                  amux = 3'b100;
                  bmux = 3'b100;
                  zmux = 3'b010;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b1;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd12:    begin
                  // aux1 <- aux1 + aux2
                  nextState = 5'd13;
                  amux = 3'b100;
                  bmux = 3'b101;
                  zmux = 3'b010;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b1;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd13:    begin
                  // aux1 <- carOscPhase + aux1
                  nextState = 5'd14;
                  amux = 3'b010;
                  bmux = 3'b100;
                  zmux = 3'b010;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b1;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd14:    begin
                  // carOscPhase <- carOscPhase + carOscDelta
                  nextState = 5'd15;
                  amux = 3'b010;
                  bmux = 3'b010;
                  zmux = 3'b010;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b1;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd15:    begin
                  // aux1 <- sine(aux1) (1)
                  nextState = 5'd16;
                  amux = 3'b100;
                  bmux = 3'bxxx;
                  zmux = 3'bxxx;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd16:    begin
                  // aux1 <- sine(aux1) (2)
                  nextState = 5'd17;
                  amux = 3'bxxx;
                  bmux = 3'bxxx;
                  zmux = 3'b100;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b1;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd17:    begin
                  // aux2 <- carEnvPhase
                  nextState = 5'd18;
                  amux = 3'b011;
                  bmux = 3'bxxx;
                  zmux = 3'b000;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b1;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd18:    begin
                  // carEnvPhase <- carEnvPhase + carEnvDelta
                  // (but only if controlTick is active AND
                  // integer part of envelope phase < 15)
                  nextState = 5'd19;
                  amux = 3'b011;
                  bmux = 3'b011;
                  zmux = 3'b010;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = controlTick &
                                    (a[15:12] != 4'd15);
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd19:    begin
                  // aux2 <- exp(fract(aux2)) (1)
                  // mulShift <- integ(aux2)
                  nextState = 5'd20;
                  amux = 3'b101;
                  bmux = 3'bxxx;
                  zmux = 3'b000;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b1;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd20:    begin
                  // aux2 <- exp(fract(aux2)) (2)
                  nextState = 5'd21;
                  amux = 3'bxxx;
                  bmux = 3'bxxx;
                  zmux = 3'b101;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b1;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd21:    begin
                  // aux1 <- (aux1 * aux2) >> (mulShift + 15) (1)
                  nextState = 5'd22;
                  amux = 3'b100;
                  bmux = 3'b101;
                  zmux = 3'bxxx;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd22:    begin
                  // aux1 <- (aux1 * aux2) >> (mulShift + 15) (2)
                  nextState = 5'd23;
                  amux = 3'bxxx;
                  bmux = 3'bxxx;
                  zmux = 3'b011;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b1;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
      5'd23:    begin
                  // signedTotal <- signedTotal + aux1
                  nextState = 5'd24;
                  amux = 3'b100;
                  bmux = 3'bxxx;
                  zmux = 3'b000;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b1;
                  loadSample = 1'b0;
                end
      5'd24:    begin
                  // sample <- unsigned(signedTotal)
                  // goto state 0
                  nextState = 5'd0;
                  amux = 3'bxxx;
                  bmux = 3'bxxx;
                  zmux = 3'bxxx;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b1;
                end
      default:  begin
                  // all other states
                  nextState = 5'd0;
                  amux = 3'bxxx;
                  bmux = 3'bxxx;
                  zmux = 3'bxxx;
                  loadModOscPhase = 1'b0;
                  loadModEnvPhase = 1'b0;
                  loadCarOscPhase = 1'b0;
                  loadCarEnvPhase = 1'b0;
                  loadAux1 = 1'b0;
                  loadAux2 = 1'b0;
                  loadMulShift = 1'b0;
                  clearSignedTotal = 1'b0;
                  addToSignedTotal = 1'b0;
                  loadSample = 1'b0;
                end
    endcase
  end

  //
  // functional units
  //

  assign addValue = a + b;
  mulsh mulsh1(clk, rst, a, b, mulShift, mulValue);
  sine sine1(clk, rst, a, sineValue);
  assign afract = { a[11:0], 4'h0 };
  exp exp1(clk, rst, afract, expValue);

  //
  // bus multiplexing
  //

  always @(*) begin
    case (amux)
      3'b000:   a = modOscPhase;
      3'b001:   a = modEnvPhase;
      3'b010:   a = carOscPhase;
      3'b011:   a = carEnvPhase;
      3'b100:   a = aux1;
      3'b101:   a = aux2;
      3'b110:   a = 16'hxxxx;
      3'b111:   a = 16'hxxxx;
      default:  a = 16'hxxxx;
    endcase
  end

  always @(*) begin
    case (bmux)
      3'b000:   b = modOscDelta;
      3'b001:   b = modEnvDelta;
      3'b010:   b = carOscDelta;
      3'b011:   b = carEnvDelta;
      3'b100:   b = aux1;
      3'b101:   b = aux2;
      3'b110:   b = 16'hxxxx;
      3'b111:   b = 16'hxxxx;
      default:  b = 16'hxxxx;
    endcase
  end

  always @(*) begin
    case (zmux)
      3'b000:   z = a;
      3'b001:   z = b;
      3'b010:   z = addValue;
      3'b011:   z = mulValue;
      3'b100:   z = sineValue;
      3'b101:   z = expValue;
      3'b110:   z = 16'hxxxx;
      3'b111:   z = 16'hxxxx;
      default:  z = 16'hxxxx;
    endcase
  end

  //
  // auxiliary registers
  //

  always @(posedge clk) begin
    if (rst) begin
      modOscPhase <= 0;
      modEnvPhase <= 0;
      carOscPhase <= 0;
      carEnvPhase <= 0;
      aux1 <= 0;
      aux2 <= 0;
      mulShift <= 0;
    end else begin
      if (start) begin
        modOscPhase <= 0;
        modEnvPhase <= 0;
        carOscPhase <= 0;
        carEnvPhase <= 0;
        aux1 <= 0;
        aux2 <= 0;
        mulShift <= 0;
      end else begin
        if (loadModOscPhase) begin
          modOscPhase <= z;
        end
        if (loadModEnvPhase) begin
          modEnvPhase <= z;
        end
        if (loadCarOscPhase) begin
          carOscPhase <= z;
        end
        if (loadCarEnvPhase) begin
          carEnvPhase <= z;
        end
        if (loadAux1) begin
          aux1 <= z;
        end
        if (loadAux2) begin
          aux2 <= z;
        end
        if (loadMulShift) begin
          mulShift <= z[15:12];
        end
      end
    end
  end

  //
  // signed total sample register
  //

  always @(posedge clk) begin
    if (rst) begin
	   signedTotal <= 20'h00000;
    end else begin
      if (clearSignedTotal) begin
        signedTotal <= 20'h00000;
      end else begin
        if (addToSignedTotal) begin
          // samples from up to 16 voices can be summed here
          // signedTotal <= signedTotal +
          //                { z[15], z[15], z[15], z[15], z[15:0] };
          // if there is only a single voice, its weight should be 16
          signedTotal <= signedTotal +
                         { z[15:0], 4'b0000 };
        end
      end
    end
  end

  //
  // sample output register
  //

  assign unsignedTotal = { ~signedTotal[19],
                           signedTotal[18:0] };

  always @(posedge clk) begin
    if (rst) begin
      sample <= 12'h800;
    end else begin
      if (loadSample) begin
        sample <= unsignedTotal[19:8];
      end
    end
  end

endmodule
